
/**************************************************************************
**  (c) Copyright 2002, Andromeda Technology & Automation
** This is free software; you can redistribute it and/or modify it under the
** terms of the GNU General Public License, see the file COPYING.
***************************************************************************
** MODULE INFORMATION *
***********************
**      FILE NAME      : message.cpp
**      SYSTEM NAME    : Gnucomo - Gnu Computer Monitoring
**      VERSION NUMBER : $Revision: 1.6 $
**
**  DESCRIPTION      :  Implementation of the message handling classes
**
**  EXPORTED OBJECTS : 
**  LOCAL    OBJECTS : 
**  MODULES  USED    :
***************************************************************************
**  ADMINISTRATIVE INFORMATION *
********************************
**      ORIGINAL AUTHOR : Arjen Baart - arjen@andromeda.nl
**      CREATION DATE   : Sep 16, 2002
**      LAST UPDATE     : Jan 31, 2003
**      MODIFICATIONS   : 
**************************************************************************/

/*****************************
   $Log: message.cpp,v $
   Revision 1.6  2003/02/05 09:37:51  arjen
   Create notifications when a new package is discovered
   in a 'rpm -qa' list or when the version of a package is changed.

   Revision 1.4  2002/12/06 22:26:28  arjen
   Set the value of log.processed to FALSE when inserting a
   new log entry into the database
   When a syslog entry arrives from last year, gcm_input subtracts one from the
   year of arrival to create the year of the log entry.
   Read output from "rpm -qa" and enter packages in the parameter table.

   Revision 1.3  2002/11/09 08:04:27  arjen
   Added a reference to the GPL

   Revision 1.2  2002/11/04 10:13:36  arjen
   Use proper namespace for iostream classes

   Revision 1.1  2002/10/05 10:25:49  arjen
   Creation of gcm_input and a first approach to a web interface

*****************************/

static const char *RCSID = "$Id: message.cpp,v 1.6 2003/02/05 09:37:51 arjen Exp $";

#include "message.h"

extern bool verbose;   /*  Defined in the main application */
extern bool testmode;

/*   Utility functions   */

String SQL_Escape(String s);

/*=========================================================================
**  NAME           : operator >>
**  SYNOPSIS       : bool operator >> (message_buffer &, String &)
**  PARAMETERS     : 
**  RETURN VALUE   : True if input was available.
**
**  DESCRIPTION    : Input operator. Read the next line from the message.
**
**  VARS USED      :
**  VARS CHANGED   :
**  FUNCTIONS USED :
**  SEE ALSO       :
**  LAST MODIFIED  : Nov 04, 2002
**=========================================================================
*/

bool operator >> (message_buffer &b, String &s)
{
   bool   input_ok = false;

   if (b.next_line == b.buffer.end())
   {
      String   l;

      if (*(b.input) >> l)
      {
         b.buffer.push_back(l);

         //   next_line keeps pointing to the end.
 
         s = l;
         input_ok = true;
      }
   }
   else
   {
      s = *(b.next_line);
      b.next_line++;
      input_ok = true;
   }
   return input_ok;
}

client_message::client_message(std::istream *in, gnucomo_database db)
{
   input.from(in);
   database = db;

   hostname = "";
   mail_header   = false;
   gpg_encrypted = false;
   classification = UNKNOWN;
   certainty      = 0.0;
}

static const String syslog_date_re("[[:alpha:]]{3} [ 123][0-9] [0-9]{2}:[0-9]{2}:[0-9]{2}");
static const String mail_date_re("[[:alpha:]]{3}, [ 123]?[0-9] [[:alpha:]]{3} [0-9]{4} [0-9]{2}:[0-9]{2}:[0-9]{2} [+-][0-9]{4}");
static const String unix_date_re("[[:alpha:]]{3} [[:alpha:]]{3} [ 123][0-9] [0-9]{2}:[0-9]{2}:[0-9]{2} [0-9]{4}");

static const regex re_syslog(syslog_date_re + " [[:alnum:]]+ [[:alpha:]]+.*:.+");
static const regex re_PGP("-----BEGIN PGP MESSAGE-----");
static const regex re_dump("^ *DUMP: Date of this level");
static const regex re_accesslog("(GET|POST) .+ HTTP");
static const regex re_errorlog("^\\[" + unix_date_re + "\\] \\[(error|notice)\\] .+");
static const regex re_rpm("[[:alnum:]+-]+-[0-9][[:alnum:].-]");

static const regex re_syslog_date("[[:alpha:]]{3} [ 123][0-9] [0-9]{2}:[0-9]{2}:[0-9]{2}");
static const regex re_uxmail_from("^From - " + unix_date_re);
static const regex re_mail_From("^From:[[:blank:]]+");
static const regex re_mail_Date("^Date:[[:blank:]]+" + mail_date_re);
static const regex re_email_address("[[:alnum:]_.-]+@[[:alnum:]_.-]+");
static const regex re_email_user("[[:alnum:]_.-]+@");

/*=========================================================================
**  NAME           : classify
**  SYNOPSIS       : double classify(String host, date arriv_d, hour arriv_t, String serv)
**  PARAMETERS     : 
**  RETURN VALUE   : The certainty with which the message is classified.
**
**  DESCRIPTION    : 
**
**  VARS USED      :
**  VARS CHANGED   :
**  FUNCTIONS USED :
**  SEE ALSO       :
**  LAST MODIFIED  : Nov 16, 2002
**=========================================================================
*/

double client_message::classify(String host, UTC arriv, String serv)
{
   String    line;

   hostname    = host;
   arrival     = arriv;
   service     = serv;

   /*  First, check if the message has a mail header. */

   if (input >> line && line == re_uxmail_from)
   {
      String    from_address;

      mail_header = true;

      /*  Scan ahead for the hostname and date of arrival.  */

      while (input >> line && line != "")
      {
         if (line == re_mail_From)
         {
            from_address = line(re_email_address);
            from_address(re_email_user) = "";            //  Remove the user part;
            hostname = from_address;
         }
         if (line == re_mail_Date)
         {
            arrival = UTC(line(regex(mail_date_re)));
         }
      }
   }
   else
   {
      //  Push the first line back, we need to read it again.
      --input;
   }

   /*
    *  Now that we have the mail header out of the way, try to figure
    *  out what the content of the message is.
    */


   while (input >> line && certainty < 0.9)
   {
      std::cout << "  testing: " << line << "\n";
      if (line == re_syslog)
      {
         certainty = 1.0;
         classification = SYSLOG;
         if (verbose)
         {
            std::cout << "Syslog detected.\n";
         }
      }
      else if (line == re_PGP)
      {
         certainty = 1.0;
         gpg_encrypted = true;
         std::cerr << "The message is PGP/GnuPG encrypted.\n";
      }
      else if (line == re_dump)
      {
          certainty = 1.0;
          if (verbose)
          {
             std::cout << "DUMP output detected.\n";
          }
      }
      else if (line == re_accesslog)
      {
          certainty = 1.0;
          classification = ACCESSLOG;
          service = "httpd";
          if (verbose)
          {
             std::cout << "HTTP access log detected.\n";
          }
      }
      else if (line == re_errorlog)
      {
          certainty = 1.0;
          classification = ERRORLOG;
          service = "httpd";
          if (verbose)
          {
             std::cout << "HTTP error log detected.\n";
          }
      }
      else if (line == re_rpm)
      {
          certainty = 1.0;
          classification = RPMLIST;
          service = "";
          if (verbose)
          {
             std::cout << "RPM package list detected.\n";
          }
      }
   }
   input.rewind();

   if (hostname == "")
   {
      std::cerr <<  "Can not determine the hostname where the message came from.\n";
      certainty = 0.0;
   }
   else if (!arrival.proper())
   {
      std::cerr << "Arrival time is not knwon.\n";
      certainty = 0.0;
   }
   else
   {
      certainty = 1.0;
   }

   return certainty;
}

/*=========================================================================
**  NAME           : enter
**  SYNOPSIS       : int enter()
**  PARAMETERS     : 
**  RETURN VALUE   : The number of lines successfully parsed from the input
**
**  DESCRIPTION    : 
**
**  VARS USED      :
**  VARS CHANGED   :
**  FUNCTIONS USED :
**  SEE ALSO       :
**  LAST MODIFIED  : Jan 31, 2003
**=========================================================================
*/

int client_message::enter()
{
   long   nr_lines = 0;
   String line;

   String change_notification("");
   String create_notification("");


   /*  Double-check the classification of the message */

   if (classification == UNKNOWN || certainty < 0.9 || gpg_encrypted)
   {
      return 0;
   }

   if (mail_header)
   {
      //  Skip the mail header.
 
      while (input >> line && line != "");
   }

   /*  Try to find the host in the database */

   String objectid;

   objectid = database.find_host(hostname);
   if (objectid == "")
   {
      std::cerr << "Please define the host " << hostname << " in the database.\n";
      return 0;
   }
   if (verbose)
   {
      std::cout << "Object id for " << hostname << " is " << objectid << "\n";
   }

   /*  Scan the input line by line, entring records into the database */

   String rest;   //  Rest of the line to be parsed

   while (input >> line)
   {
      if (verbose)
      {
         std::cout << line << "\n";
      }


      /*  Check each line if it contains valid information */

      const regex *check;

      switch (classification)
      {
      case SYSLOG:
            check = &re_syslog;
            break;
      case ACCESSLOG:
            check = &re_accesslog;
            break;
      case ERRORLOG:
            check = &re_errorlog;
            break;
      case RPMLIST:
            check = &re_rpm;
            break;
      }

      if (line == *check)
      {
         date   log_date;
         hour   log_time;
         int    i;

         String insertion("insert into log (objectid, servicecode,"
                           " object_timestamp, timestamp, rawdata, processed) values (");
         String datestring;

         switch (classification)
         {
         case SYSLOG:
            log_date = line;
            log_time = line;
            if (log_date.Year() < 0 || log_date.Year() > 2500)
            {
               //  The year is not in the log file. Assume the year of arrival,
               //  unless this puts the log entry at a later date than the arrival date.
               //  This happens e.g. when a log entry from December arrives in Januari.

               log_date = date(log_date.Day(), log_date.Month(), date(arrival).Year());
               if (log_date > date(arrival))
               {
                  log_date = date(log_date.Day(), log_date.Month(), date(arrival).Year() - 1);
               }
            }

            if (verbose)
            {
               std::cout << "   Log timestamp  = " << log_date << " " << log_time << "\n";
            }
            rest = line << 16;
            i = rest.index(' ');
            if (rest(0,i) == hostname(0,i))
            {
               rest <<= i + 1;
               if (verbose)
               {
                  std::cout << "   Hostname matches.\n";
                  std::cout << "   rest = " << rest << "\n";
               }
               for (i = 0; isalpha(rest[i]) && i < ~rest; i++);
               if (verbose)
               {
                  std::cout << "   Service name = " << rest(0,i) << "\n";
               }

               /*   Insert a new record into the log table   */

               insertion += "'" + objectid + "',";
               insertion += "'" + rest(0,i) + "',";
               insertion += "'" + log_date.format("%Y-%m-%d") + " " + log_time.format() + "',";
               insertion += "'" + arrival.format("%Y-%m-%d %T") + "',";
               insertion += "'" + SQL_Escape(line) + "',FALSE";
               insertion += ")";
            
               if (testmode)
               {
                  std::cout << insertion << "\n";
               }
               else
               {
                  database.Query(insertion);
               }

               if (verbose)
               {
                  std::cout << "\n\n";
               }

               nr_lines++;
            }
            else
            {
               std::cerr << "   Hostname " << rest(0,i) << " does not match.\n";
            }
            break;

         case ACCESSLOG:
            datestring = line(regex("\\[.+\\]"));
            datestring <<= 1;
            datestring >>= 1;
            datestring[datestring.index(':')] = ' ';
            log_date = datestring;
            log_time = datestring;
            if (verbose)
            {
               std::cout << "   Log timestamp  = " << log_date << " " << log_time << "\n";
            }
            insertion += "'" + objectid + "',";
            insertion += "'" + service + "',";
            insertion += "'" + log_date.format("%Y-%m-%d") + " " + log_time.format() + "',";
            insertion += "'" + arrival.format("%Y-%m-%d %T") + "',";
            insertion += "'" + SQL_Escape(line) + "',FALSE";
            insertion += ")";
            
            if (testmode)
            {
               std::cout << insertion << "\n";
            }
            else
            {
               database.Query(insertion);
            }

            if (verbose)
            {
               std::cout << "\n\n";
            }

            nr_lines++;
            break;

         case ERRORLOG:
            datestring = line(regex("\\[.+\\]"));
            datestring <<= 1;
            datestring >>= 1;
            log_date = datestring;
            log_time = datestring;
            if (verbose)
            {
               std::cout << "   Log timestamp  = " << log_date << " " << log_time << "\n";
            }
            insertion += "'" + objectid + "',";
            insertion += "'" + service + "',";
            insertion += "'" + log_date.format("%Y-%m-%d") + " " + log_time.format() + "',";
            insertion += "'" + arrival.format("%Y-%m-%d %T") + "',";
            insertion += "'" + SQL_Escape(line) + "',FALSE";
            insertion += ")";
            
            if (testmode)
            {
               std::cout << insertion << "\n";
            }
            else
            {
               database.Query(insertion);
            }

            if (verbose)
            {
               std::cout << "\n\n";
            }

            nr_lines++;
            break;

         case RPMLIST:
            //  Scan a list of packages and versions from "rpm -a".
	    //  A similar listing can be created on IRIX 6.5 by using the
	    //  command "showprods -3 -n|awk '{printf "%s-%s\n",$2,$3}'|grep -v '^[-=]' \
	    //            |grep -v Version-Description".
	    //
            //  We have to separate the package name and the version.
            //  The separation is marked by a '-', followed by a digit.
 
            String qry = "select count(paramid) from parameter where objectid='";
            qry += objectid + "' and class='package'";

            database.Query(qry);
            long n_packages = String(database.Field(0, "count"));

            bool initial_entry = n_packages == 0;

            i = line.index('-');
            while (!(line[i] == '-' && isdigit(line[i + 1])))
            {
               i++;
            }
            String package(line(0,i));
            String version(line(i+1, ~line));
            String paramid;
            String remark;
            String insert_h;

            if (verbose)
            {
               std::cout << "Package is " << package;
               std::cout << ", version is " << version << "\n";
            }

            //  Construct a qry to check the package's existance

            qry = "select paramid from parameter where objectid='";
            qry += objectid + "' and class='package' and name='";
            qry += package + "'";

            if (database.Query(qry) == 1)
            {
               paramid = database.Field(0, "paramid");
               qry = "select value from property where paramid='";
               qry += paramid + "' and name='version'";
               if (database.Query(qry) == 0)
               {
                  std::cerr << "Database corruption: Package " << package;
                  std::cerr << " does not have a 'version' property.\n";
               }
               else if (database.Field(0, "value") != version)
               {
                  if (verbose)
                  {
                     std::cout << "  Parameter " << package << " has different version\n";
                  }
                  insertion = "update property set value='";
                  insertion += version + "' where paramid='";
                  insertion += paramid + "' and name='version'";

                  insert_h = "insert into history (paramid, modified, change_nature, changed_property, new_value)";
                  insert_h += " values ('";
                  insert_h += paramid + "', '" + arrival.format("%Y-%m-%d %T") + "', 'MODIFIED', 'version', '";
                  insert_h += version + "')";

                  database.Query(insertion);
                  database.Query(insert_h);

                  if (change_notification == "")
                  {
                     remark = "Gnucomo detected a different version for package parameter(s) ";
                     change_notification = database.new_notification(objectid, "property modified", remark);
                  }

                  insertion = "insert into parameter_notification (notificationid, paramid) values ('";
                  insertion += change_notification + "', '";
                  insertion += paramid + "')";

                  database.Query(insertion);
               }
               else
               {
                  if (verbose)
                  {
                     std::cout << "   Parameter " << package << " has not changed.\n";
                  }
               }
            }
            else
            {

               if (verbose)
               {
                  std::cout << "  Parameter " << package << " does not exist.\n";
               }
               //  Create a new package parameter, including version property and history record

               insertion = "insert into parameter (objectid, name, class, description) values ('";
               insertion += objectid + "', '" + package + "', 'package', 'RPM package " + package + "')";
               if (testmode)
               {
                  paramid = "0";
                  std::cout << insertion << "\n";
               }
               else
               {
                  database.Query(insertion);
                  qry = "select paramid from parameter where objectid='";
                  qry += objectid + "' and class='package' and name='";
                  qry += package + "'";
                  database.Query(qry);
                  paramid = database.Field(0, "paramid");
               }

               insertion = "insert into property (paramid, name, value, type) values ('";
               insertion += paramid + "', 'version', '";
               insertion += version + "', 'STATIC')";
               insert_h = "insert into history (paramid, modified, change_nature, changed_property, new_value)";
               insert_h += " values ('";
               insert_h += paramid + "', '" + arrival.format("%Y-%m-%d %T") + "', 'CREATED', 'version', '";
               insert_h += version + "')";

               if (testmode)
               {
                  std::cout << insertion << "\n" << insert_h << "\n";
               }
               else
               {
                  database.Query(insertion);
                  database.Query(insert_h);
                  if (!initial_entry)
                  {
                     if (create_notification == "")
                     {
                        remark = "Gnucomo detected new parameter(s) of class package";
                        create_notification = database.new_notification(objectid, "parameter created", remark);
                     }
                     insertion = "insert into parameter_notification (notificationid, paramid) values ('";
                     insertion += create_notification + "', '";
                     insertion += paramid + "')";

                     database.Query(insertion);
                  }
               }
            }

            if (verbose)
            {
               std::cout << "\n";
            }

            nr_lines++;
            break;

         }
      }
      else
      {
         std::cerr << "gcm_input WARNING: Not a valid line: " << line << "\n";
      }
   }

   if (verbose)
   {
      std::cout << nr_lines << " lines parsed from the log file.\n";
   }
   return nr_lines;
}

/*=========================================================================
**  NAME           : SQL_Escape
**  SYNOPSIS       : String SQL_Escape(String)
**  PARAMETERS     : 
**  RETURN VALUE   : 
**
**  DESCRIPTION    : Insert backslashes before single quotes.
**
**  VARS USED      :
**  VARS CHANGED   :
**  FUNCTIONS USED :
**  SEE ALSO       :
**  LAST MODIFIED  : 
**=========================================================================
*/

String SQL_Escape(String s)
{
   int i;

   for (i = 0; i < ~s; i++)
   {
      if (s[i] == '\'')
      {
         s(i,0) = "\\";
         i++;
      }
   }

   return s;
}
